# ZenML constants
TEST_RESULTS = 'test_results'

# TFX constants
RUN_FN = 'run_fn'
CUSTOM_CONFIG = 'custom_config'
EXAMPLES = 'examples'
SCHEMA = 'schema'
BASE_MODEL = 'base_model'
TRANSFORM_GRAPH = 'transform_graph'
HYPERPARAMETERS = 'hyperparameters'
MODEL = 'model'
MODEL_RUN = 'model_run'
