#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pandas Datasource definition"""

from typing import Text

from zenml.datasources import BaseDatasource


class PandasDatasource(BaseDatasource):
    """ZenML Pandas datasource definition."""

    def __init__(
            self,
            name: Text,
            df,
            **kwargs):
        """
        Initialize pandas datasource.
        Args:
            name: Name of datasource.
            df: pandas dataframe
        """
        super().__init__(name, **kwargs)
        raise NotImplementedError('Its coming soon!')
